"""
The Klein Bottle manifold.

Lead author: Juliane Braunsmann
"""

import geomstats.backend as gs
from geomstats.geometry.manifold import Manifold
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.vectorization import get_batch_shape, repeat_out


class KleinBottle(Manifold):
    r"""Class for the Klein Bottle, a two dimensional manifold which is not orientable.

    Points are understood to be on the Klein Bottle if they are in the interval
    :math:`[0,1]^2`.
    Each point in :math:`\mathbb R^2` can be understood as a point on the Klein Bottle
    by considering the equivalence relation

    .. math::
       :nowrap:

       \begin{align}
         &  (x_1,y_1) \sim (x_2, y_2) \\
         \Leftrightarrow \quad &  y_1-y_2 \in \mathbb Z
                                 \text{ and } x_1-x_2 \in 2\mathbb Z \\
         \text{ or } \quad &
        y_1 + y_2 \in \mathbb Z \text{ and } x_1-x_2 \in \mathbb Z\setminus 2\mathbb Z.
       \end{align}
    """

    def __init__(self, equip=True):
        super().__init__(dim=2, shape=(2,), intrinsic=True, equip=equip)

    def default_metric(self):
        """Metric to equip the space with if equip is True."""
        return KleinBottleMetric

    def random_point(self, n_samples=1, bound=None):
        """Uniformly sample points on the manifold.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : unused

        Returns
        -------
        samples : array-like, shape=[n_samples, 2]
            Points sampled on the manifold.
        """
        samples = gs.random.uniform(size=(n_samples, 2))
        if n_samples == 1:
            return gs.squeeze(samples, axis=0)
        return samples

    def to_tangent(self, vector, base_point=None):
        """Project a vector to a tangent space of the manifold.

        In this case returns the vector itself.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : unused

        Returns
        -------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at base point.
        """
        if not gs.all(self.is_tangent(vector)):
            raise ValueError("Cannot handle non tangent vectors")

        tangent_vec = gs.copy(vector)
        return repeat_out(
            self.point_ndim, tangent_vec, vector, base_point, out_shape=self.shape
        )

    def is_tangent(self, vector, base_point=None, atol=gs.atol):
        r"""Evaluate if the point belongs to :math:`\mathbb R^2`.

        This method checks the shape of the input point.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Point to test.
        base_point: unused here
        atol : float
            Unused here.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the tangent space.
        """
        belongs = self.shape == vector.shape[-self.point_ndim :]
        shape = get_batch_shape(self.point_ndim, vector, base_point)
        if belongs:
            return gs.ones(shape, dtype=bool)
        return gs.zeros(shape, dtype=bool)

    def belongs(self, point, atol=gs.atol):
        """Evaluate if the point belongs to the set [0,1]^2.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point to test.
        atol : float
            Unused here.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the space.
        """
        return gs.all(gs.logical_and(point >= 0, point <= 1), axis=-1)

    @staticmethod
    def equivalent(point_a, point_b, atol=gs.atol):
        r"""Evaluate whether two points represent equivalent points on the Klein Bottle.

        This method uses the equivalence stated in the class description.

        This means points are equivalent if one walks an even number of squares in
        x-direction and any number of squares in y-direction or an uneven number of
        squares in x-direction and any number of squares in y-direction while
        "mirroring" the y coordinate.

        Parameters
        ----------
        point_a: array-like, shape=[..., 2]
            Representation of point on Klein Bottle
        point_b: array-like, shape=[..., 2]
            Representation of point on Klein Bottle
        atol: Absolute tolerance to test for belonging to \mathbb Z.

        Returns
        -------
        is_equivalent : array-like, shape=[..., 2]
            Boolean evaluating if points are equivalent
        """
        x_diff = point_a[..., 0] - point_b[..., 0]
        y_diff = point_a[..., 1] - point_b[..., 1]
        y_sum = point_a[..., 1] + point_b[..., 1]
        un_mirrored = gs.logical_and(
            _is_close_mod(gs.mod(x_diff, 2.0), 2.0, atol=atol),
            _is_close_mod(gs.mod(y_diff, 1.0), 1.0, atol=atol),
        )
        mirrored = gs.logical_and(
            _is_close_mod(gs.mod(x_diff - 1, 2.0), 2.0, atol=atol),
            _is_close_mod(gs.mod(y_sum, 1.0), 1.0, atol=atol),
        )
        return gs.logical_or(un_mirrored, mirrored)

    def regularize(self, point):
        r"""Regularize arbitrary point to canonical equivalent in unit square.

        Regularize any point in :math:`\mathbb R^2` to its canonical equivalent
        representation in the square :math:`[0,1]^2`.

        Parameters
        ----------
        point : array-like, shape=[..., 2]
            Point.

        Returns
        -------
        regularized_point : array-like, shape=[..., 2]
            Regularized point.
        """
        # determine number of steps to take in x direction
        num_steps = gs.cast(gs.abs(gs.floor(point[..., [0]])), gs.int64)

        x_canonical = gs.mod(point[..., 0], 1.0)
        y_canonical_even = gs.mod(point[..., 1], 1.0)
        y_canonical_odd = gs.mod(1 - y_canonical_even, 1.0)

        point_even = gs.stack([x_canonical, y_canonical_even], axis=-1)
        point_odd = gs.stack([x_canonical, y_canonical_odd], axis=-1)
        return gs.where(gs.mod(num_steps, 2) == 0, point_even, point_odd)

    def to_coords(self, point, coords_type, **kwargs):
        """Convert point from intrinsic to coordinates type.

        Parameters
        ----------
        point : array-like, shape=[..., 2]
            Point on the Klein bottle, in intrinsic coordinates.
        coords_type : str
            One of the following: "extrinsic", "bottle", "bagel"

        Returns
        -------
        transformed_point : array-like, shape=[..., repr_dim]
            Point in new representation.
        """
        coords_func = getattr(self, f"to_{coords_type}_coords")
        return coords_func(point, **kwargs)

    @staticmethod
    def to_extrinsic_coords(point):
        r"""Convert point to extrinsic coordinates.

        Convert from the intrinsic coordinates in the Klein bottle (2 parameters),
        to the extrinsic coordinates in Euclidean space (4 parameters).
        For intrinsic parameters (\theta,v) the extrinsic Euclidean parametrization is
        [https://en.wikipedia.org/wiki/Klein_bottle#4-D_non-intersecting]:

        x = R \left(cos(\theta/2)cos(v) - sin(\theta/2)sin(2v)\right)
        y = R \left(sin(\theta/2)cos(v) + cos(\theta/2)sin(2v)\right)
        z = P cos(\theta)(1+\epsilon sin(v))
        w = P sin(\theta)(1+\epsilon sin(v))

        for 0\leq\theta<2\pi and 0\leq v<2\pi. P and R are constants to determine
        the aspect ratio.
        ε is any small constant .

        Parameters
        ----------
        point : array-like, shape=[..., 2]
            Point on the Klein bottle, in intrinsic coordinates.

        Returns
        -------
        point_extrinsic : array-like, shape=[..., 4]
            Point on the Klein bottle, in extrinsic coordinates in
            Euclidean space.
        """
        theta = 2 * gs.pi * point[..., 0]
        v = 2 * gs.pi * point[..., 1]
        R = 1
        P = 1
        epsilon = 0.1

        x = R * (gs.cos(theta / 2) * gs.cos(v) - gs.sin(theta / 2) * gs.sin(2 * v))
        y = R * (gs.sin(theta / 2) * gs.cos(v) + gs.cos(theta / 2) * gs.sin(2 * v))
        z = P * gs.cos(theta) * (1 + epsilon * gs.sin(v))
        w = P * gs.sin(theta) * (1 + epsilon * gs.sin(v))

        return gs.stack([x, y, z, w], axis=-1)

    @staticmethod
    def to_bottle_coords(point):
        r"""Convert point to coordinates in R^3 parametrizing the Klein bottle.

        Convert from the intrinsic coordinates in the Klein bottle (2 parameters),
        to the coordinates of the Klein bottle parametrization in 3d Euclidean space
        (3 parameters).
        For intrinsic parameters (\theta,v) the Klein bottle parametrization is
        [https://en.wikipedia.org/wiki/Klein_bottle#Bottle_shape]

        Parameters
        ----------
        point : array-like, shape=[..., 2]
            Point on the Klein bottle, in intrinsic coordinates.

        Returns
        -------
        point_extrinsic : array-like, shape=[..., 4]
            Point on the Klein bottle, in the Klein bagel parametrization.
        """
        u = 2 * gs.pi * point[..., 0]
        v = 2 * gs.pi * point[..., 1]

        fx1, fx2, fx3, fx4, fx5, fx6 = [2 / 15, 3, 30, 90, 60, 5]

        x = (
            -fx1
            * gs.cos(u)
            * (
                fx2 * gs.cos(v)
                - fx3 * gs.sin(u)
                + fx4 * (gs.cos(u)) ** 4 * gs.sin(u)
                - fx5 * (gs.cos(u)) ** 6 * gs.sin(u)
                + fx6 * gs.cos(u) * gs.cos(v) * gs.sin(u)
            )
        )

        fy1, fy2, fy3, fy4, fy5, fy6, fy7, fy8, fy9, fy10 = [
            1 / 15,
            3,
            3,
            48,
            48,
            60,
            5,
            5,
            80,
            80,
        ]

        y = (
            -fy1
            * gs.sin(u)
            * (
                fy2 * gs.cos(v)
                - fy3 * (gs.cos(u)) ** 2 * gs.cos(v)
                - fy4 * (gs.cos(u)) ** 4 * gs.cos(v)
                + fy5 * (gs.cos(u)) ** 6 * gs.cos(v)
                - fy6 * gs.sin(u)
                + fy7 * gs.cos(u) * gs.cos(v) * gs.sin(u)
                - fy8 * (gs.cos(u)) ** 3 * gs.cos(v) * gs.sin(u)
                - fy9 * (gs.cos(u)) ** 5 * gs.cos(v) * gs.sin(u)
                + fy10 * (gs.cos(u)) ** 10 * gs.cos(v) * gs.sin(u)
            )
        )

        fz1, fz2, fz3 = [2 / 15, 3, 5]
        z = fz1 * (fz2 + fz3 * gs.cos(u) * gs.sin(u)) * gs.sin(v)

        return gs.stack([x, y, z], axis=-1)

    @staticmethod
    def to_bagel_coords(point, radius=5.0):
        r"""Convert point to coordinates in R^3 parametrizing the Klein bagel.

        Convert from the intrinsic coordinates in the Klein bottle (2 parameters),
        to the coordinates of the Klein bagel parametrization in 3d Euclidean space
        (3 parameters).
        For intrinsic parameters (\theta,v) the Klein bagel parametrization is
        [https://en.wikipedia.org/wiki/Klein_bottle#The_figure_8_immersion]:

        x = \left(r + cos(\theta/2)sin(v) - sin(\theta/2)sin(2v)\right)cos(theta)
        y = \left(r + cos(\theta/2)sin(v) - sin(\theta/2)sin(2v)\right)sin(theta)
        z = sin(\theta/2)sin(v) + cos(\theta/2)sin(2v)

        for 0\leq\theta<2\pi and 0\leq v<2\pi. r is a constant to determine
        the aspect ratio.

        Parameters
        ----------
        point : array-like, shape=[..., 2]
            Point on the Klein bottle, in intrinsic coordinates.
        radius : float
            Radius of the circle.

        Returns
        -------
        point_extrinsic : array-like, shape=[..., 4]
            Point on the Klein bottle, in the Klein bagel parametrization.
        """
        theta = 2 * gs.pi * point[..., 0]
        v = 2 * gs.pi * point[..., 1]

        x = (
            radius + gs.cos(theta / 2) * gs.sin(v) - gs.sin(theta / 2) * gs.sin(2 * v)
        ) * gs.cos(theta)
        y = (
            radius + gs.cos(theta / 2) * gs.sin(v) - gs.sin(theta / 2) * gs.sin(2 * v)
        ) * gs.sin(theta)
        z = gs.sin(theta / 2) * gs.sin(v) + gs.cos(theta / 2) * gs.sin(2 * v)

        return gs.stack([x, y, z], axis=-1)


class KleinBottleMetric(RiemannianMetric):
    """Class for the Klein Bottle Metric.

    Implements exp and log using explicit formulas.

    Parameters
    ----------
    space : KleinBottle
        Underlying manifold.
    """

    def injectivity_radius(self, base_point=None):
        """Compute the radius of the injectivity domain.

        This is is the supremum of radii r for which the exponential map is a
        diffeomorphism from the open ball of radius r centered at the base
        point onto its image.

        Parameters
        ----------
        base_point : array-like, shape=[..., 2]
            Point on the manifold. Unused.

        Returns
        -------
        radius : float
            Injectivity radius.
        """
        radius = gs.array(0.5)
        return repeat_out(self._space.point_ndim, radius, base_point)

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point=None):
        """Inner product between two tangent vectors at a base point.

        Parameters
        ----------
        tangent_vec_a: array-like, shape=[..., 2]
            Tangent vector at base point.
        tangent_vec_b: array-like, shape=[..., 2]
            Tangent vector at base point.
        base_point: array-like, shape=[..., 2]
            Base point.
            Optional, default: None, unused.

        Returns
        -------
        inner_product : array-like, shape=[...,]
            Inner-product.
        """
        return gs.dot(tangent_vec_a, tangent_vec_b)

    def exp(self, tangent_vec, base_point):
        """Exponential map.

        Computed by adding tangent_vec to base_point and finding canonical
        representation in the unit square.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., 2]
            Tangent vector at the base point.
        base_point : array-like, shape=[..., 2]
            Point on the manifold.

        Returns
        -------
        exp : array-like, shape=[..., dim]
            Point on the manifold.
        """
        if base_point.ndim == 1 and tangent_vec.ndim > 1:
            while base_point.ndim < tangent_vec.ndim:
                base_point = gs.expand_dims(base_point, 0)
        base_point_canonical = self._space.regularize(base_point)
        point = base_point_canonical + tangent_vec
        return self._space.regularize(point)

    def log(self, point, base_point):
        """Logarithm map.

        Computed by finding the representative of point closest to base_point and
        returning their difference.

        Parameters
        ----------
        point : array-like, shape=[..., 2]
            Point on the manifold.
        base_point : array-like, shape=[..., 2]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at the base point.
        """
        if base_point.ndim == 1 and point.ndim > 1:
            while base_point.ndim < point.ndim:
                base_point = gs.expand_dims(base_point, 0)
        base_point_canonical, point_minimal = self._closest_representative(
            base_point, point
        )
        return point_minimal - base_point_canonical

    def _closest_representative(self, point1, point2):
        """Find the representative of point2 which is closest to point1.

        Only representatives in the 8 surrounding
        squares have to be considered.

        Parameters
        ----------
        point1: array-like, shape [..., 2]
            Point on the manifold.
        point2: array-like, shape [..., 2]
            Point on the manifold.

        Returns
        -------
        point1: array-like, shape [...,2]
            Canonical representation of point1 in unit square.
        minimizers: array-like, shape [...,2]
            Representation of point2 with smallest distance to point1.
        """
        point1, point2 = gs.broadcast_arrays(point1, point2)
        shape = point1.shape
        p1 = self._space.regularize(point1)
        p1 = gs.reshape(p1, (-1, 2))
        p2 = self._space.regularize(point2)
        p2 = gs.reshape(p2, (-1, 2))
        p2_2 = gs.stack([p2[:, 0], p2[:, 1] - 1], axis=-1)
        p2_3 = gs.stack([p2[:, 0], p2[:, 1] + 1], axis=-1)
        p2_4 = gs.stack([p2[:, 0] + 1, 1 - p2[:, 1]], axis=-1)
        p2_5 = gs.stack([p2[:, 0] + 1, 2 - p2[:, 1]], axis=-1)
        p2_6 = gs.stack([p2[:, 0] + 1, 0 - p2[:, 1]], axis=-1)
        p2_7 = gs.stack([p2[:, 0] - 1, 1 - p2[:, 1]], axis=-1)
        p2_8 = gs.stack([p2[:, 0] - 1, 2 - p2[:, 1]], axis=-1)
        p2_9 = gs.stack([p2[:, 0] - 1, 0 - p2[:, 1]], axis=-1)
        p2_total = gs.stack(
            [p2, p2_2, p2_3, p2_4, p2_5, p2_6, p2_7, p2_8, p2_9], axis=0
        )
        indices = gs.argmin(
            gs.sum((p2_total - gs.expand_dims(p1, 0)) ** 2, axis=-1), axis=0
        )
        minimizers = gs.empty_like(p1)
        for i, index in enumerate(indices):
            minimizers[i, :] = p2_total[index, i, :]
        p1 = gs.reshape(p1, shape)
        minimizers = gs.reshape(minimizers, shape)
        return p1, minimizers


def _is_close_mod(array, divisor, atol):
    """Determine if values in array are elementwise close to zero modulo divisor.

    Evaluate elementwise to true if the absolute value is close to 0 or divisor.

    Parameters
    ----------
    array: array-like, shape [...]
        Array who values will be considered.
    divisor: int
        Divisor in the modulo operation.
    atol: float
        Absolute tolerance threshold.

    Return
    ------
    _is_close_mod_bool: boolean array-like, shape [...]
        Elementwise result.
    """
    return gs.logical_or(
        gs.isclose(gs.abs(array), 0.0, atol), gs.isclose(gs.abs(array), divisor, atol)
    )



"""Transformer for manifold-valued data.

Lead author: Nicolas Guigui.
"""

from sklearn.base import BaseEstimator, TransformerMixin

import geomstats.backend as gs
from geomstats.geometry.matrices import Matrices
from geomstats.geometry.skew_symmetric_matrices import SkewSymmetricMatrices
from geomstats.geometry.symmetric_matrices import SymmetricMatrices
from geomstats.learning.exponential_barycenter import ExponentialBarycenter
from geomstats.learning.frechet_mean import FrechetMean


class ToTangentSpace(BaseEstimator, TransformerMixin):
    """Lift data to a tangent space.

    Compute the logs of all data points and reshape them to
    1d vectors if necessary. This means that all the data points, that belong
    to a possibly non-linear manifold are lifted to one of the tangent space of
    the manifold, which is a vector space. By default, the mean of the data
    is computed (with the FrechetMean or the ExponentialBarycenter estimator,
    as appropriate) and the tangent space at the mean is used. Any other base
    point can be passed. The data points are then represented by the initial
    velocities of the geodesics that lead from base_point to each data point.
    Any machine learning algorithm can then be used with the output array.

    Parameters
    ----------
    space : Manifold
        Equipped manifold or unequipped space implementing `exp` and `log`.

    Notes
    -----
    * Required geometry methods: `log`, `exp`.
    """

    def __init__(self, space):
        self.space = space

        if hasattr(self.space, "metric"):
            self.mean_estimator = FrechetMean(space)
        else:
            self.mean_estimator = ExponentialBarycenter(space)

        self.base_point_ = None

    @property
    def _geometry(self):
        """Object where `exp` and `log` are defined."""
        if hasattr(self.space, "metric"):
            return self.space.metric

        return self.space

    def fit(self, X, y=None, weights=None, base_point=None):
        """Compute the central point at which to take the log.

        This method is only used if `base_point=None` to compute the mean of
        the input data.

        Parameters
        ----------
        X : array-like, shape=[n_samples, {dim, [n, n]}]
            The training input samples.
        y : None
            Ignored.
        weights : array-like, shape=[n_samples, 1]
            Weights associated to the points.
            Optional, default: None
        base_point : array-like, shape=[{dim, [n, n]}]
            Point similar to the input data from which to compute the logs.
            Optional, default: None.

        Returns
        -------
        self : object
            Returns self.
        """
        if base_point is None:
            self.base_point_ = self.mean_estimator.fit(X, y, weights).estimate_

        else:
            self.base_point_ = base_point

        return self

    def transform(self, X):
        """Lift data to a tangent space.

        Compute the logs of all data point and reshapes them to
        1d vectors if necessary. By default the logs are taken at the mean
        but any other base point can be passed. Any machine learning
        algorithm can then be used with the output array.

        Parameters
        ----------
        X : array-like, shape=[n_samples, {dim, [n, n]}]
            Data to transform.

        Returns
        -------
        X_new : array-like, shape=[n_samples, dim]
            Lifted data.
        """
        if self.base_point_ is None:
            raise Exception("Not fitted")

        tangent_vecs = self._geometry.log(X, base_point=self.base_point_)

        if self.space.point_ndim == 1:
            return tangent_vecs

        if gs.all(Matrices.is_symmetric(tangent_vecs)):
            X = SymmetricMatrices.basis_representation(tangent_vecs)

        elif gs.all(Matrices.is_skew_symmetric(tangent_vecs)):
            X = SkewSymmetricMatrices(tangent_vecs.shape[-1]).basis_representation(
                tangent_vecs
            )
        else:
            X = gs.reshape(tangent_vecs, (len(X), -1))

        return X

    def inverse_transform(self, X):
        """Reconstruction of X.

        The reconstruction will match X_original whose transform would be X.

        Parameters
        ----------
        X : array-like, shape=[n_samples, dim]
            New data, where dim is the dimension of the manifold data belong
            to.

        Returns
        -------
        X_original : array-like, shape=[n_samples, {dim, [n, n]}
            Data lying on the manifold.
        """
        if self.base_point_ is None:
            raise Exception("Not fitted")

        if self.space.point_ndim > 1:
            n_base_point = self.base_point_.shape[-1]
            n_vecs = X.shape[-1]
            dim_sym = int(n_base_point * (n_base_point + 1) / 2)
            dim_skew = int(n_base_point * (n_base_point - 1) / 2)

            if gs.all(Matrices.is_symmetric(self.base_point_)) and dim_sym == n_vecs:
                tangent_vecs = SymmetricMatrices(
                    self.base_point_.shape[-1]
                ).matrix_representation(X)
            elif dim_skew == n_vecs:
                tangent_vecs = SkewSymmetricMatrices(dim_skew).matrix_representation(X)
            else:
                dim = self.base_point_.shape[-1]
                tangent_vecs = gs.reshape(X, (len(X), dim, dim))
        else:
            tangent_vecs = X
        return self._geometry.exp(tangent_vecs, self.base_point_)



"""The Siegel manifold.

The Siegel manifold is a generalization of the Poincare disk to complex matrices
of singular values strictly lower than one.
It is defined as the set of complex matrices M such that:
I - M @ M.conj().T is a positive definite matrix.
Warning: another more restrictive definition of the Siegel disk also exists
which add a symmetry condition on the matrices.
It has been proven in [Cabanes2022]_ that the sub-manifold of symmetric Siegel
matrices is a totally geodesic sub-manifold of the Siegel space.
The sub-manifold of real Siegel matrices is also a totally geodesic sub-manifold
of the Siegel space.

Lead author: Yann Cabanes.

References
----------
.. [Cabanes2022] Yann Cabanes. Multidimensional complex stationary
    centered Gaussian autoregressive time series machine learning
    in Poincaré and Siegel disks: application for audio and radar
    clutter classification, PhD thesis, 2022
.. [Cabanes2021] Yann Cabanes and Frank Nielsen.
    New theoreticla tools in the Siegel space for vectorial
    autoregressive data classification,
    Geometric Science of Information, 2021.
    https://franknielsen.github.io/IG/GSI2021-SiegelLogExpClassification.pdf
.. [JV2016] B. Jeuris and R. Vandebril. The Kahler mean of Block-Toeplitz
    matrices with Toeplitz structured blocks, 2016.
    https://epubs.siam.org/doi/pdf/10.1137/15M102112X
"""

import geomstats.backend as gs
from geomstats.geometry.base import ComplexVectorSpaceOpenSet
from geomstats.geometry.complex_matrices import ComplexMatrices
from geomstats.geometry.complex_riemannian_metric import ComplexRiemannianMetric
from geomstats.geometry.hermitian_matrices import expmh, powermh
from geomstats.geometry.matrices import Matrices


class Siegel(ComplexVectorSpaceOpenSet):
    """Class for the Siegel space.

    Parameters
    ----------
    n : int
        Integer representing the shape of the matrices: n x n.
    symmetric : bool
        If symmetric is True, add a symmetry condition
        on the matrices to belong to the Siegel space.
        Optional, default: False.
    """

    def __init__(self, n, symmetric=False, equip=True):
        super().__init__(
            dim=n**2,
            embedding_space=ComplexMatrices(m=n, n=n),
            equip=equip,
        )
        self.n = n
        self.symmetric = symmetric

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return SiegelMetric

    def belongs(self, point, atol=gs.atol):
        """Check if a matrix belongs to the Siegel space.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Point to be checked.
        atol : float
            Tolerance.
            Optional, default: backend atol.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean denoting if mat belongs to the Siegel space.
        """
        point_transconj = ComplexMatrices.transconjugate(point)
        aux = gs.matmul(point, point_transconj)

        axis = -1 if gs.ndim(point) == 3 else None
        belongs = gs.all(gs.linalg.eigvalsh(aux) <= 1 + atol, axis=axis)

        if self.symmetric:
            return gs.logical_and(belongs, Matrices.is_symmetric(point))

        return belongs

    def projection(self, point, atol=gs.atol):
        """Project a matrix to the Siegel space.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Matrix to project.
        atol : float
            Tolerance.
            Optional, default: backend atol.

        Returns
        -------
        projected: array-like, shape=[..., n, n]
            Matrix in the Siegel space.
        """
        if self.symmetric:
            point = Matrices.to_symmetric(point)

        point_transconj = ComplexMatrices.transconjugate(point)
        aux = gs.matmul(point, point_transconj)

        eigenvalues = gs.linalg.eigvalsh(aux)
        max_eigenvalues = gs.amax(eigenvalues, axis=-1) ** 0.5

        scalars = gs.where(
            (max_eigenvalues > 1.0 - gs.atol), (1 - atol) / max_eigenvalues, 1.0
        )
        return gs.einsum("...,...ij->...ij", scalars, point)

    def random_point(self, n_samples=1, bound=1.0):
        """Generate random points in the Siegel space.

        The Siegel space is the set of complex matrices of singular values
        strictly lower than one.
        The Frobenius norm of a matrix is greater than or equal to the spectral norm
        which corresponds to the largest singular value of a matrix.
        Then, simulating a matrix with Frobenius norm strictly lower than one,
        its singular values are also strictly lower than one,
        therefore this matrix belongs to the Siegel disk.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Bound of the interval in which to sample in the tangent space.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n, n]
            Points sampled in the Siegel space.
        """
        n = self.n
        size = (n_samples, n, n) if n_samples != 1 else (n, n)

        samples = gs.random.rand(*size, dtype=gs.get_default_cdtype())
        samples -= 0.5 + 0.5j
        samples *= bound * (1 - gs.atol) * 2**0.5 / n
        return samples

    def random_tangent_vec(self, base_point, n_samples=1):
        """Sample on the tangent space of Siegel space from the uniform distribution.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        base_point : array-like, shape=[..., n, n]
            Base point of the tangent space.
            Optional, default: None.

        Returns
        -------
        samples : array-like, shape=[..., n, n]
            Points sampled in the tangent space at base_point.
        """
        n = self.n
        size = (n_samples, n, n) if n_samples != 1 else (n, n)

        samples = gs.random.rand(*size, dtype=gs.get_default_cdtype())
        samples *= 2
        samples -= 1 + 1j

        return self.to_tangent(samples, base_point)


class SiegelMetric(ComplexRiemannianMetric):
    """Class for the Siegel metric."""

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point):
        """Compute the Siegel inner-product.

        Compute the inner-product of tangent_vec_a and tangent_vec_b
        at point base_point using the Siegel Riemannian metric.
        The expression of the inner product between the vectors `v` and `w`
        at point `O` is :math:`<v, w>_{O}
        = 1/2 * trace((I - O O^{H})^{-1} v (I - O^{H} O)^{-1} w^{H})
        + 1/2 * trace((I - O O^{H})^{-1} w (I - O^{H} O)^{-1} v^{H})
        = Re(trace((I - O O^{H})^{-1} v (I - O^{H} O)^{-1} w^{H}))`
        where H denotes the conjugate transpose operator.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n, n]
            Tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        inner_product : array-like, shape=[...,]
            Inner-product.
        """
        identity = gs.eye(base_point.shape[-1], dtype=base_point.dtype)

        base_point_transconj = ComplexMatrices.transconjugate(base_point)
        tangent_vec_b_transconj = ComplexMatrices.transconjugate(tangent_vec_b)

        aux_1 = gs.matmul(base_point, base_point_transconj)

        aux_2 = gs.matmul(base_point_transconj, base_point)

        aux_3 = identity - aux_1

        aux_4 = identity - aux_2

        inv_aux_3 = powermh(aux_3, -1)

        inv_aux_4 = powermh(aux_4, -1)

        aux_a = gs.matmul(inv_aux_3, tangent_vec_a)
        aux_b = gs.matmul(inv_aux_4, tangent_vec_b_transconj)
        trace = Matrices.trace_product(aux_a, aux_b)

        return gs.real(trace)

    @staticmethod
    def tangent_vec_from_base_point_to_zero(tangent_vec, base_point):
        """Transport a tangent vector from a base point to zero.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at zero.
        base_point : array-like, shape=[..., n, n]
            Point on the Siegel space.

        Returns
        -------
        tangent_vec_at_zero : array-like, shape=[..., n, n]
            Tangent vector at zero (null matrix).
        """
        identity = gs.eye(base_point.shape[-1], dtype=base_point.dtype)
        base_point_transconj = ComplexMatrices.transconjugate(base_point)
        aux_1 = gs.matmul(base_point, base_point_transconj)
        aux_2 = gs.matmul(base_point_transconj, base_point)
        aux_3 = identity - aux_1
        aux_4 = identity - aux_2
        factor_1 = powermh(aux_3, -1 / 2)
        factor_3 = powermh(aux_4, -1 / 2)
        prod_1 = gs.matmul(factor_1, tangent_vec)
        return gs.matmul(prod_1, factor_3)

    @staticmethod
    def exp_at_zero(tangent_vec):
        """Compute the Siegel exponential map at zero.

        Compute the exponential map at zero (null matrix) of tangent_vec.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.

        Returns
        -------
        exp : array-like, shape=[..., n, n]
            Point on the manifold.
        """
        identity = gs.eye(tangent_vec.shape[-1], dtype=tangent_vec.dtype)
        tangent_vec_transconj = ComplexMatrices.transconjugate(tangent_vec)
        aux_1 = gs.matmul(tangent_vec, tangent_vec_transconj)
        aux_2 = powermh(aux_1, 1 / 2)
        aux_3 = expmh(2 * aux_2)
        factor_1 = aux_3 - identity
        aux_4 = aux_3 + identity
        factor_2 = powermh(aux_4, -1)
        factor_3 = powermh(aux_2, -1)
        factor_3 = gs.where(gs.isnan(factor_3), gs.zeros_like(factor_2), factor_3)
        prod_1 = gs.matmul(factor_1, factor_2)
        prod_2 = gs.matmul(prod_1, factor_3)
        return gs.matmul(prod_2, tangent_vec)

    @staticmethod
    def isometry(point, point_to_zero):
        """Define an isometry for the Siegel metric.

        Isometry for the Siegel metric sending point_to_zero
        (parameter of the isometry) on zero.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Point on the Siegel space.
        point_to_zero : array-like, shape=[..., n, n]
            Point send on zero (null matrix) by the isometry.

        Returns
        -------
        point_image : array-like, shape=[..., n, n]
            Image of point by the isometry.
        """
        identity = gs.eye(point.shape[-1], dtype=point.dtype)
        point_to_zero_transconj = ComplexMatrices.transconjugate(point_to_zero)
        aux_1 = gs.matmul(point_to_zero, point_to_zero_transconj)
        aux_2 = gs.matmul(point_to_zero_transconj, point_to_zero)
        aux_3 = identity - aux_1
        aux_4 = identity - aux_2
        factor_1 = powermh(aux_3, -1 / 2)
        factor_4 = powermh(aux_4, 1 / 2)
        factor_2 = point - point_to_zero
        aux_5 = gs.matmul(point_to_zero_transconj, point)
        aux_6 = identity - aux_5
        factor_3 = gs.linalg.inv(aux_6)
        prod_1 = gs.matmul(factor_1, factor_2)
        prod_2 = gs.matmul(prod_1, factor_3)
        return gs.matmul(prod_2, factor_4)

    def exp(self, tangent_vec, base_point):
        """Compute the Siegel exponential map.

        Compute the exponential map at base_point of tangent_vec.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Point on the manifold.

        Returns
        -------
        exp : array-like, shape=[..., n, n]
            Point on the manifold.
        """
        tangent_vec_at_zero = self.tangent_vec_from_base_point_to_zero(
            tangent_vec=tangent_vec, base_point=base_point
        )

        exp_zero = self.exp_at_zero(tangent_vec_at_zero)

        return self.isometry(point=exp_zero, point_to_zero=-base_point)

    @staticmethod
    def log_at_zero(point):
        """Compute the Siegel logarithm map at zero.

        Compute the logarithm map at zero (null matrix) of point.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Point on the Siegel space.

        Returns
        -------
        log_at_zero : array-like, shape=[..., n, n]
            Riemannian logarithm at zero (null matrix).
        """
        identity = gs.eye(point.shape[-1], dtype=point.dtype)
        point_transconj = ComplexMatrices.transconjugate(point)
        aux_1 = gs.matmul(point, point_transconj)
        aux_2 = powermh(aux_1, 1 / 2)
        num = identity + aux_2
        den = identity - aux_2
        inv_den = powermh(den, -1)
        frac = gs.matmul(num, inv_den)
        factor_1 = gs.linalg.logm(frac)
        factor_2 = powermh(aux_2, -1)
        factor_2 = gs.where(gs.isnan(factor_2), gs.zeros_like(factor_2), factor_2)
        prod_1 = gs.matmul(factor_1, factor_2)
        return gs.matmul(prod_1, point) * 0.5

    @staticmethod
    def tangent_vec_from_zero_to_base_point(tangent_vec, base_point):
        """Transport a tangent vector from zero to a base point.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at zero.
        base_point : array-like, shape=[..., n, n]
            Point on the Siegel space.

        Returns
        -------
        tangent_vec_at_base_point : array-like, shape=[..., n, n]
            Tangent vector at the base point.
        """
        identity = gs.eye(base_point.shape[-1], dtype=base_point.dtype)
        base_point_transconj = ComplexMatrices.transconjugate(base_point)
        aux_1 = gs.matmul(base_point, base_point_transconj)
        aux_2 = gs.matmul(base_point_transconj, base_point)
        aux_3 = identity - aux_1
        aux_4 = identity - aux_2
        factor_1 = powermh(aux_3, 1 / 2)
        factor_3 = powermh(aux_4, 1 / 2)
        prod_1 = gs.matmul(factor_1, tangent_vec)
        return gs.matmul(prod_1, factor_3)

    def log(self, point, base_point):
        """Compute the Siegel logarithm map.

        Compute the Riemannian logarithm at point base_point
        of point wrt the metric defined in inner_product.
        Return a tangent vector at point base_point.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Point on the Siegel space.
        base_point : array-like, shape=[..., n, n]
            Point on the Siegel space.

        Returns
        -------
        log : array-like, shape=[..., n, n]
            Riemannian logarithm at the base point.
        """
        point_at_zero = self.isometry(point=point, point_to_zero=base_point)

        logarithm_at_zero = self.log_at_zero(point_at_zero)

        return self.tangent_vec_from_zero_to_base_point(
            tangent_vec=logarithm_at_zero, base_point=base_point
        )

    def squared_dist(self, point_a, point_b):
        """Compute the Siegel squared distance.

        Compute the Riemannian squared distance between point_a and point_b.

        Parameters
        ----------
        point_a : array-like, shape=[..., n, n]
            Point.
        point_b : array-like, shape=[..., n, n]
            Point.

        Returns
        -------
        squared_dist : array-like, shape=[...]
            Riemannian squared distance.
        """
        if gs.ndim(point_a) > gs.ndim(point_b):
            point_a, point_b = point_b, point_a

        identity = gs.eye(point_a.shape[-1], dtype=point_a.dtype)

        point_a_transconj = ComplexMatrices.transconjugate(point_a)
        point_b_transconj = ComplexMatrices.transconjugate(point_b)

        factor_1 = point_b - point_a

        factor_3 = ComplexMatrices.transconjugate(factor_1)

        aux_2 = gs.matmul(point_a_transconj, point_b)

        aux_3 = gs.matmul(point_a, point_b_transconj)

        aux_4 = identity - aux_2

        aux_5 = identity - aux_3

        factor_2 = gs.linalg.inv(aux_4)

        factor_4 = gs.linalg.inv(aux_5)

        prod = gs.einsum(
            "...ij,...jk,...kl,...lm->...im", factor_1, factor_2, factor_3, factor_4
        )

        prod_power_one_half = gs.linalg.fractional_matrix_power(prod, 0.5)

        num = identity + prod_power_one_half
        den = identity - prod_power_one_half

        inv_den = gs.linalg.inv(den)

        frac = gs.matmul(num, inv_den)

        logarithm = gs.linalg.logm(frac)

        sq_dist = Matrices.trace_product(logarithm, logarithm) * 0.25
        sq_dist = gs.real(sq_dist)
        return gs.maximum(sq_dist, 0)

    def sectional_curvature_at_zero(self, tangent_vec_a, tangent_vec_b, atol=gs.atol):
        """Compute the sectional curvature at zero.

        Non-orthonormal vectors can be given.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n, n]
            Tangent vector at zero.
        tangent_vec_b : array-like, shape=[..., n, n]
            Tangent vector at zero.
        atol : float
            Tolerance.
            Optional, default: backend atol.

        Returns
        -------
        sectional_curvature : array-like, shape=[...,]
            Sectional curvature at zero.
        """

        def _scale_by_norm(tangent_vec):
            norm_tangent_vec = self.norm(tangent_vec, base_point=zero)
            scalars = gs.where(norm_tangent_vec < atol, 0.0, 1 / norm_tangent_vec)
            return _scale(scalars, tangent_vec)

        def _scale(scalars, tangent_vec):
            return gs.einsum("...,...ij->...ij", scalars, tangent_vec)

        zero = gs.zeros([self._space.n, self._space.n], dtype=tangent_vec_a.dtype)

        tangent_vec_a = _scale_by_norm(tangent_vec_a)

        inner_prod = gs.cast(
            self.inner_product(tangent_vec_a, tangent_vec_b, base_point=zero),
            dtype=tangent_vec_a.dtype,
        )

        tangent_vec_b = tangent_vec_b - _scale(inner_prod, tangent_vec_a)
        tangent_vec_b = _scale_by_norm(tangent_vec_b)

        tangent_vec_a_transconj = ComplexMatrices.transconjugate(tangent_vec_a)
        tangent_vec_b_transconj = ComplexMatrices.transconjugate(tangent_vec_b)

        term1 = gs.matmul(tangent_vec_a, tangent_vec_b_transconj)
        term1 -= gs.matmul(tangent_vec_b, tangent_vec_a_transconj)
        norm_term1 = gs.linalg.norm(term1, axis=(-2, -1)) ** 2

        term2 = gs.matmul(tangent_vec_a_transconj, tangent_vec_b)
        term2 -= gs.matmul(tangent_vec_b_transconj, tangent_vec_a)
        norm_term2 = gs.linalg.norm(term2, axis=(-2, -1)) ** 2
        return -0.5 * (norm_term1 + norm_term2)

    def sectional_curvature(
        self, tangent_vec_a, tangent_vec_b, base_point=None, atol=gs.atol
    ):
        """Compute the sectional curvature.

        For two orthonormal tangent vectors at a base point :math:`x,y`,
        the sectional curvature is defined by :math:`<R(x, y)x,
        y>`. Non-orthonormal vectors can be given.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n, n]
            Tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point. Optional, default is zero
        atol : float
            Tolerance.
            Optional, default: backend atol.

        Returns
        -------
        sectional_curvature : array-like, shape=[...,]
            Sectional curvature at `base_point`.
        """
        if base_point is not None:
            tangent_vec_a = self.tangent_vec_from_base_point_to_zero(
                tangent_vec_a, base_point
            )
            tangent_vec_b = self.tangent_vec_from_base_point_to_zero(
                tangent_vec_b, base_point
            )
        return self.sectional_curvature_at_zero(tangent_vec_a, tangent_vec_b, atol=atol)



from geomstats.test_cases.learning._base import BaseEstimatorTestCase


class RiemannianMinimumDistanceToMeanTestCase(BaseEstimatorTestCase):
    def test_fit(self, X_train, y_train, expected, atol):
        mean_estimates = self.estimator.fit(X_train, y_train).mean_estimates_
        self.assertAllClose(mean_estimates, expected, atol=atol)

    def test_predict(self, X_train, y_train, X_test, y_test, atol):
        self.estimator.fit(X_train, y_train)

        y_pred = self.estimator.predict(X_test)
        self.assertAllClose(y_pred, y_test, atol=atol)

    def test_predict_proba(self, X_train, y_train, X_test, expected_proba, atol):
        self.estimator.fit(X_train, y_train)

        proba = self.estimator.predict_proba(X_test)
        self.assertAllClose(proba, expected_proba, atol=atol)

    def test_transform(self, X_train, y_train, X_test, expected, atol):
        self.estimator.fit(X_train, y_train)

        dists = self.estimator.transform(X_test)
        self.assertAllClose(dists, expected, atol=atol)

    def test_score(self, X_train, y_train, X_test, y_expected, atol):
        self.estimator.fit(X_train, y_train)

        score = self.estimator.score(X_test, y_expected)
        self.assertAllClose(score, 1.0, atol=atol)



"""Module providing an implementation of MatrixLieAlgebras.

There are two main forms of representation for elements of a MatrixLieAlgebra
implemented here. The first one is as a matrix, as elements of R^(n x n).
The second is by choosing a base and remembering the coefficients of an element
in that base. This base will be provided in child classes
(e.g. SkewSymmetricMatrices).

Lead author: Stefan Heyder.
"""

import abc

import geomstats.backend as gs
import geomstats.errors
from geomstats.geometry.base import MatrixVectorSpace
from geomstats.geometry.matrices import Matrices

from ._bch_coefficients import BCH_COEFFICIENTS


class MatrixLieAlgebra(MatrixVectorSpace, abc.ABC):
    """Class implementing matrix Lie algebra related functions.

    Parameters
    ----------
    representation_dim : int
        Amount of rows and columns in the matrix representation of the
        Lie algebra.
    """

    # TODO: check again need for representation_dim

    def __init__(self, representation_dim, **kwargs):
        geomstats.errors.check_integer(representation_dim, "representation_dim")
        super().__init__(shape=(representation_dim, representation_dim), **kwargs)
        self.representation_dim = representation_dim

    bracket = Matrices.bracket

    def baker_campbell_hausdorff(self, matrix_a, matrix_b, order=2):
        """Calculate the Baker-Campbell-Hausdorff approximation of given order.

        The implementation is based on [CM2009a]_ with the pre-computed
        constants taken from [CM2009b]_. Our coefficients are truncated to
        enable us to calculate BCH up to order 15.

        This represents Z = log(exp(X)exp(Y)) as an infinite linear combination
        of the form Z = sum z_i e_i where z_i are rational numbers and e_i are
        iterated Lie brackets starting with e_1 = X, e_2 = Y, each e_i is given
        by some i',i'': e_i = [e_i', e_i''].

        Parameters
        ----------
        matrix_a : array-like, shape=[..., *point_shape]
        matrix_b : array-like, shape=[..., *point_shape]
            Matrices.
        order : int
            The order to which the approximation is calculated. Note that this
            is NOT the same as using only e_i with i < order.
            Optional, default 2.

        References
        ----------
        .. [CM2009a] F. Casas and A. Murua. An efficient algorithm for
            computing the Baker–Campbell–Hausdorff series and some of its
            applications. Journal of Mathematical Physics 50, 2009
        .. [CM2009b] http://www.ehu.eus/ccwmuura/research/bchHall20.dat
        """
        if order > 15:
            raise NotImplementedError("BCH is not implemented for order > 15.")

        number_of_hom_degree = gs.array(
            [2, 1, 2, 3, 6, 9, 18, 30, 56, 99, 186, 335, 630, 1161, 2182]
        )
        n_terms = gs.sum(number_of_hom_degree[:order])

        el = [matrix_a, matrix_b]
        result = matrix_a + matrix_b

        for i in gs.arange(2, n_terms):
            i_p = BCH_COEFFICIENTS[i, 1] - 1
            i_pp = BCH_COEFFICIENTS[i, 2] - 1

            el.append(self.bracket(el[i_p], el[i_pp]))
            result += (
                float(BCH_COEFFICIENTS[i, 3]) / float(BCH_COEFFICIENTS[i, 4]) * el[i]
            )
        return result



import geomstats.backend as gs
from geomstats.test.random import RandomDataGenerator
from geomstats.test.test_case import TestCase


class FiniteDifferenceTestCase(TestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.space)

    def _linear_path(self, base_point, tangent_vec, time):
        ijk = "ijk"[: self.space.point_ndim]
        if base_point.shape[: -self.space.point_ndim]:
            base_point = gs.expand_dims(base_point, axis=-(self.space.point_ndim + 1))

        return base_point + gs.einsum(f"t,...{ijk}->...t{ijk}", time, tangent_vec)



"""Geomstats custom exceptions."""


class AutodiffNotImplementedError(RuntimeError):
    """Raised when autodiff is not implemented."""


class NotPartialOrder(Exception):
    """Raise an exception when less equal is not true."""
